/* pnmctrfilt.c - apply centre filter transform to a portable pixmap
**
** Copyright (C) 2003 by John Walker
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.

** The source code for this program is in the public domain.  Executable
** versions are subject to the licenses of all libraries they include.
*/

#include <math.h>
#include <ctype.h>
#include "pnm.h"
#include "shhopt.h"

struct cmdlineInfo {
    /* All the information the user supplied in the command line,
       in a form easy for the program to use.
    */
    const char *filespec;  /* '-' if stdin */
    double power, density, radius;
    int xshift, yshift;
};


static void
parseCommandLine(int argc, char ** argv, 
                 struct cmdlineInfo * const cmdlineP) {

    optEntry *option_def = malloc(100 * sizeof(optEntry));
        /* Instructions to optParseOptions3 on how to parse our options. */
    optStruct3 opt;

    unsigned int option_def_index;
    float fpow, fdens, frad;

    option_def_index = 0;   /* incremented by OPTENT3 */

    OPTENT3(0, "power",    OPT_FLOAT,   &fpow,              NULL, 0);
    OPTENT3(0, "density",  OPT_FLOAT,   &fdens,             NULL, 0);
    OPTENT3(0, "radius",   OPT_FLOAT,   &frad,              NULL, 0);
    OPTENT3(0, "xshift",   OPT_INT,     &cmdlineP->xshift,  NULL, 0);
    OPTENT3(0, "yshift",   OPT_INT,     &cmdlineP->yshift,  NULL, 0);
    
    fpow = 1.0;
    fdens = 2.0;
    cmdlineP->xshift = 0;
    cmdlineP->yshift = 0;
    cmdlineP->radius = 1.0;

    opt.opt_table = option_def;
    opt.short_allowed = FALSE;  /* We have no short (old-fashioned) options */
    opt.allowNegNum = TRUE; 

    optParseOptions3(&argc, argv, opt, sizeof(opt), 0);
        /* Uses and sets argc, argv, and some of *cmdline_p and others. */
	
    cmdlineP->power = fpow;
    cmdlineP->density = fdens;
    cmdlineP->radius = frad;

#ifdef CFDEBUG    
    fprintf(stderr, "Power = %g   Density = %g, Radius = %g\n", cmdlineP->power, cmdlineP->density, cmdlineP->radius);
#endif

    if (argc == 1) {
        cmdlineP->filespec = "-";
    } else if (argc == 2) {
    	cmdlineP->filespec = argv[1];
    } else {
        pm_error("Wrong number of arguments.  You may have at most one filename");
    }

}


/*----------------------------------------------------------------------------
  
   Build a gamma table of size maxval+1 for the given gamma value.
  
   This function depends on pow(3m).  If you don't have it, you can
   simulate it with '#define pow(x,y) exp((y)*log(x))' provided that
   you have the exponential function exp(3m) and the natural logarithm
   function log(3m).  I can't believe I actually remembered my log
   identities.
-----------------------------------------------------------------------------*/


int
main(int argc, char *argv[] ) {
    struct cmdlineInfo cmdline;
    FILE* ifp;
    xel* xelrow;
    xelval maxval;
    int rows, cols, format, row;
    int i, xsize, ysize, diagonal, erad, xctr, yctr;
    int outputFormat;
    double *ldens;

    pnm_init( &argc, argv );

    parseCommandLine(argc, argv, &cmdline);

    ifp = pm_openr(cmdline.filespec);

    pnm_readpnminit(ifp, &cols, &rows, &maxval, &format);
    xelrow = pnm_allocrow(cols);

    if (PNM_FORMAT_TYPE(format) == PPM_TYPE)
        outputFormat = PPM_TYPE;
    else 
        outputFormat = PGM_TYPE;

    if (PNM_FORMAT_TYPE(format) != outputFormat) {
        if (outputFormat == PGM_TYPE)
            pm_message("Promoting to PGM");
    }

    pnm_writepnminit( stdout, cols, rows, maxval, outputFormat, 0 );
    
    /* Determine the radius of the filter.  This is the half diagonal
       measure of the image multiplied by the command line radius factor. */
    
    xsize = (cols + 1) / 2;
    ysize = (rows + 1) / 2;
    erad = (int) (sqrt((xsize * xsize) + (ysize * ysize)) + 0.5) * cmdline.radius;
#ifdef CFDEBUG    
    fprintf(stderr, "Image size: %d by %d, filter radius = %d\n", cols, rows, erad);
#endif

    /* Build the in-memory table which maps distance from the
       centre of the image (as adjusted by the X and Y offset,
       if any) to the density of the filter at this remove.  This
       table needs to be as large as the diagonal from the
       (possibly offset) centre to the most distant corner
       of the image. */

    xsize = ((cols + 1) / 2) + ABS(cmdline.xshift);
    ysize = ((rows + 1) / 2) + ABS(cmdline.yshift);
    diagonal = ((int) (sqrt((xsize * xsize) + (ysize * ysize)) + 0.5)) +  1;
    
    ldens = malloc(diagonal * sizeof(double));
    for (i = 0; i < diagonal; i++) {
    	/* If you don't have the "pow()" function, you can define it as:
	    	#define pow(x, y) (exp((y) * log(x)))
	*/
    	ldens[i] = (i >= erad) ? 1 :
	    (1.0 + (cmdline.density - 1) * pow(1.0 - (((double) i) / (erad - 1)), cmdline.power));
/*fprintf(stderr, "Density at %d is %g\n", i, ldens[i]); */
    }
    
    xctr = ((cols + 1) / 2) + cmdline.xshift;
    yctr = ((rows + 1) / 2) + cmdline.yshift;
    
    for (row = 0; row < rows; ++row) {
        int col;
	int xd, yd = ABS(yctr - row), td;
	
        pnm_readpnmrow(ifp, xelrow, cols, maxval, format);

        pnm_promoteformatrow(xelrow, cols, maxval, format, 
                             maxval, outputFormat);

        switch (PNM_FORMAT_TYPE(outputFormat)) {
            case PPM_TYPE:
        	for (col = 0; col < cols; ++col) {
                    xelval r, g, b;

    	    	    xd = ABS(xctr - col);
		    td = (int) (sqrt((xd * xd) + (yd * yd)) + 0.5);
                    r = (int) (PPM_GETR(xelrow[col]) / ldens[td]);
                    g = (int) (PPM_GETG(xelrow[col]) / ldens[td]);
                    b = (int) (PPM_GETB(xelrow[col]) / ldens[td]);
                    PPM_ASSIGN(xelrow[col], r, g, b);
        	}
        	break;

            case PGM_TYPE:
        	for (col = 0; col < cols; ++col) {
                    xelval xel;

    	    	    xd = ABS(xctr - col);
		    td = (int) (sqrt((xd * xd) + (yd * yd)) + 0.5);
/*fprintf(stderr, "(%d,%d) [%d,%d]: %d\n", col, row, xd, yd, td);*/
		    xel = (int) (PNM_GET1(xelrow[col]) / ldens[td]);
                    PNM_ASSIGN1(xelrow[col], xel);
        	}
        	break;
            default:
        	pm_error("Internal error");
        }
        pnm_writepnmrow(stdout, xelrow, cols, maxval, outputFormat, 0);
    }

    pm_close(ifp);
    pm_close(stdout);

    return 0;
}

